# SPDX-License-Identifier: GPL-3.0-or-later

from pathlib import Path

from gi.repository import Gtk

from .global_state import global_state
from .installation_scripting import installation_scripting, Step
from .page import Page
from .widgets import reset_model, DeviceRow


@Gtk.Template(resource_path='/com/github/p3732/os-installer/ui/pages/welcome.ui')
class WelcomePage(Gtk.Box, Page):
    __gtype_name__ = __qualname__
    image = 'weather-clear-symbolic'

    description = Gtk.Template.Child()

    def __init__(self, **kwargs):
        Gtk.Box.__init__(self, **kwargs)
        config = global_state.get_config('welcome_page')

        if config['logo']:
            self.image = Path(config['logo'])

    ### callbacks ###

    @Gtk.Template.Callback('continue')
    def _continue(self, button):
        global_state.advance(self)

    ### public methods ###

    def load_once(self):
        config = global_state.get_config('welcome_page')
        language_code = global_state.get_config('language_code')

        if (text_key := f'text_{language_code}') in config:
            text = config[text_key]
        elif config['text']:
            text = config['text']
        else:
            text = self.description.get_label()
            text = text.format(global_state.get_config('distribution_name'))
        self.description.set_label(text)



# SPDX-License-Identifier: GPL-3.0-or-later

from threading import Lock

from gi.repository import Gio, Gtk

from .disk_provider import disk_provider
from .global_state import global_state
from .installation_scripting import installation_scripting, Step
from .page import Page
from .system_calls import is_booted_with_uefi, open_disks
from .widgets import reset_model, DeviceRow

GIGABYTE_FACTOR = 1000 * 1000 * 1000


@Gtk.Template(resource_path='/com/github/p3732/os-installer/ui/pages/disk.ui')
class DiskPage(Gtk.Stack, Page):
    __gtype_name__ = __qualname__
    no_disk_image_name = 'no-disk-symbolic'
    default_image_name = 'drive-harddisk-system-symbolic'
    image = default_image_name
    can_reload = True

    disk_label = Gtk.Template.Child()
    disk_size = Gtk.Template.Child()
    disk_list = Gtk.Template.Child()

    whole_disk_row = Gtk.Template.Child()
    partition_stack = Gtk.Template.Child()
    partition_list = Gtk.Template.Child()

    disk_list_model = Gio.ListStore()
    partition_list_model = Gio.ListStore()

    current_disk = None
    lock = Lock()

    def __init__(self, **kwargs):
        Gtk.Stack.__init__(self, **kwargs)

        self.minimum_disk_size = global_state.get_config('minimum_disk_size') * GIGABYTE_FACTOR

        # models
        self.disk_list.bind_model(self.disk_list_model, self._create_device_row)
        self.partition_list.bind_model(self.partition_list_model, self._create_device_row)

    def _create_device_row(self, info):
        if info.size <= 0 or info.size >= self.minimum_disk_size:
            return DeviceRow(info)
        else:
            required_size_str = disk_provider.disk_size_to_str(self.minimum_disk_size)
            return DeviceRow(info, required_size_str)

    def _setup_disk_list(self):
        if global_state.demo_mode:
            disks = disk_provider.get_testing_dummy_disks()
        else:
            disks = disk_provider.get_disks()

        if len(disks) == 0:
            self.set_visible_child_name('no-disks')
            self.image = self.no_disk_image_name
        else:
            reset_model(self.disk_list_model, disks)
            self.set_visible_child_name('disks')
            self.image = self.default_image_name
        global_state.reload_title_image()

    def _setup_partition_list(self, disk_info):
        self.current_disk = disk_info

        # set disk info
        self.disk_label.set_label(disk_info.name)
        self.whole_disk_row.set_subtitle(disk_info.device_path)
        self.disk_size.set_label(disk_info.size_text)

        # reset partition list
        if len(disk_info.partitions) == 0:
            reset_model(self.partition_list_model, [])
            self.partition_stack.set_visible_child_name("no-partitions")
        elif is_booted_with_uefi() and disk_info.efi_partition is None:
            self.partition_stack.set_visible_child_name("no-boot-partition")
        else:
            reset_model(self.partition_list_model, disk_info.partitions)
            self.partition_stack.set_visible_child_name("available")

        # show
        self.set_visible_child_name('partitions')

    def _store_device_info(self, info):
        global_state.set_config('disk_name', info.name)
        global_state.set_config('disk_device_path', info.device_path)
        global_state.set_config('disk_is_partition', not type(info) == type(self.current_disk))
        global_state.set_config('disk_efi_partition', self.current_disk.efi_partition)

    ### callbacks ###

    @Gtk.Template.Callback('clicked_disks_button')
    def _clicked_disks_button(self, button):
        open_disks()

    @Gtk.Template.Callback('disk_selected')
    def _disk_selected(self, list_box, row):
        with self.lock:
            self._setup_partition_list(row.info)
            self.can_navigate_backward = True

    @Gtk.Template.Callback('reload')
    def _reload(self, button):
        global_state.reload_page()

    @Gtk.Template.Callback('use_partition')
    def _use_partition(self, list_box, row):
        info = row.info
        if not info.name:
            info.name = row.get_title()
        self._store_device_info(info)
        global_state.advance(self)

    @Gtk.Template.Callback('use_whole_disk')
    def _use_whole_disk(self, list_box, row):
        self._store_device_info(self.current_disk)
        global_state.advance(self)

    ### public methods ###

    def load(self):
        if not self.loaded:
            self.loaded = True
            # start prepare script
            installation_scripting.set_ok_to_start_step(Step.prepare)

        with self.lock:
            self._setup_disk_list()

    def unload(self):
        self.can_navigate_backward = False

    def navigate_backward(self):
        with self.lock:
            self.can_navigate_backward = False
            self.set_visible_child_name('disks')



# SPDX-License-Identifier: GPL-3.0-or-later

from enum import Enum
from threading import Lock
import os

from gi.repository import Gio, GLib, Vte

from .global_state import global_state


class Step(Enum):
    none = 0
    prepare = 1
    install = 2
    configure = 3
    done = 4


class InstallationScripting():
    '''
    Handles all calls to scripts for installation. The installation process consists of 3 steps:
    * Preparation. Used e.g. for updating mirrors.
    * Installation. Installs an OS onto a disk.
    * Configuration. Configures an OS according to user's choices.
    '''

    terminal = Vte.Terminal()
    cancel = Gio.Cancellable()

    lock = Lock()
    current_step = Step.none
    step_ready = Step.none
    script_running = False

    # set by installation page
    install_page_name = None

    def __init__(self):
        # setup terminal
        self.terminal.set_input_enabled(False)
        self.terminal.set_scroll_on_output(True)
        self.terminal.set_hexpand(True)
        self.terminal.set_vexpand(True)
        self.terminal.connect('child-exited', self._on_child_exited)

    def _start_next_script(self):
        if self.current_step is Step.configure:
            global_state.installation_running = False
            global_state.advance(None, allow_return=global_state.demo_mode,
                                 cleanup=not global_state.demo_mode,)

        if self.current_step.value < self.step_ready.value and not self.script_running:
            self.current_step = Step(self.current_step.value + 1)
            print(f'Starting step "{self.current_step.name}"...')

            is_installing = self.current_step is Step.install or self.current_step is Step.configure
            is_configuring = self.current_step is Step.configure
            envs = global_state.create_envs(is_installing, is_configuring)

            global_state.installation_running = is_installing

            # check config
            if envs == None:
                print(f'Not all config options set for "{self.current_step.name}". Please report this bug.')
                print('############################')
                print(global_state.config)
                print('############################')
                global_state.installation_failed()
                return

            # start script
            file_name = f'/etc/os-installer/scripts/{self.current_step.name}.sh'
            if os.path.exists(file_name):
                self.script_running, _ = self.terminal.spawn_sync(
                    Vte.PtyFlags.DEFAULT, '/', ['sh', file_name],
                    envs, GLib.SpawnFlags.DEFAULT, None, None, self.cancel)
                if not self.script_running:
                    print(f'Could not start {self.current_step.name} script! Ignoring.')
                    self._start_next_script()
            else:
                print(f'No script for step {self.current_step.name} exists.')
                self._start_next_script()

    ### callbacks ###

    def _on_child_exited(self, terminal, status):
        with self.lock:
            self.script_running = False
            print(f'Finished step "{self.current_step.name}".')

            if not status == 0 and not global_state.demo_mode:
                global_state.installation_failed()
            else:
                self._start_next_script()

    ### public methods ###

    def set_ok_to_start_step(self, step: Step):
        with self.lock:
            if self.step_ready.value < step.value:
                self.step_ready = step
                self._start_next_script()


installation_scripting = InstallationScripting()



# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Gio, Gtk

from .global_state import global_state
from .keyboard_layout_provider import get_default_layout, get_layouts_for
from .language_provider import language_provider
from .page import Page
from .system_calls import set_system_keyboard_layout
from .widgets import reset_model, ProgressRow


@Gtk.Template(resource_path='/com/github/p3732/os-installer/ui/pages/keyboard_layout.ui')
class KeyboardLayoutPage(Gtk.Box, Page):
    __gtype_name__ = __qualname__
    image = 'input-keyboard-symbolic'

    stack = Gtk.Template.Child()

    # overview
    chosen_layouts = Gtk.Template.Child()
    continue_button = Gtk.Template.Child()
    primary_layout_row = Gtk.Template.Child()

    # layouts
    language_row = Gtk.Template.Child()
    layout_list = Gtk.Template.Child()

    # languages
    language_list = Gtk.Template.Child()

    languages_model = Gio.ListStore()
    layouts_model = Gio.ListStore()

    current_row = None
    language_list_setup = False
    loaded_language = ''

    def __init__(self, **kwargs):
        Gtk.Box.__init__(self, **kwargs)

        # models
        self.layout_list.bind_model(self.layouts_model, lambda o: ProgressRow(o.name, o))
        self.language_list.bind_model(self.languages_model, lambda o: ProgressRow(o.name, o))

    def _setup_languages_list(self):
        languages = language_provider.get_all_languages()
        reset_model(self.languages_model, languages)

    def _load_overview(self, keyboard_info):
        self.stack.set_visible_child_name('overview')
        self.primary_layout_row.set_title(keyboard_info.name)
        set_system_keyboard_layout(keyboard_info.name, keyboard_info.layout)

    def _load_layout_list(self, language, language_code):
        self.stack.set_visible_child_name('layouts')

        if self.loaded_language == language_code:
            return

        self.loaded_language = language_code
        self.language_row.set_subtitle(language)

        # fill list with all keyboard layouts for given language
        layouts = get_layouts_for(language_code, language)
        reset_model(self.layouts_model, layouts)


    ### callbacks ###

    @Gtk.Template.Callback('continue')
    def _continue(self, button):
        global_state.advance(self)

    @Gtk.Template.Callback('language_row_activated')
    def _language_row_activated(self, list_box, row):
        # show layouts for language
        language_info = row.info
        self._load_layout_list(language_info.name, language_info.language_code)

    @Gtk.Template.Callback('layout_row_activated')
    def _layout_row_activated(self, list_box, row):
        # use selected keyboard layout
        keyboard_info = row.info
        global_state.set_config('keyboard_layout', keyboard_info)
        self._load_overview(keyboard_info)
        self.can_navigate_backward = False

    @Gtk.Template.Callback('show_language_selection')
    def _show_language_selection(self, row):
        if not self.language_list_setup:
            self.language_list_setup = True
            self._setup_languages_list()

        # show language selection
        self.stack.set_visible_child_name('languages')
        self.can_navigate_backward = True

    @Gtk.Template.Callback('show_layout_selection')
    def _show_layout_selection(self, row):
        if self.loaded_language == '':
            self._load_layout_list(self.default_language, self.default_language_code)
        else:
            self.stack.set_visible_child_name('layouts')
        self.can_navigate_backward = True

    ### public methods ###

    def load_once(self):
        # page gets reconstructed if different app language is chosen
        self.default_language = global_state.get_config('language')
        self.default_language_code = global_state.get_config('language_code')
        keyboard_info = get_default_layout(self.default_language_code)
        self._load_overview(keyboard_info)

    def navigate_backward(self):
        match self.stack.get_visible_child_name():
            case 'layouts':
                self.stack.set_visible_child_name('overview')
                self.can_navigate_backward = False
            case 'languages':
                self.stack.set_visible_child_name('layouts')



# SPDX-License-Identifier: GPL-3.0-or-later

''' All calls to other programs are encapsulated here. '''

import os
from subprocess import Popen
import subprocess
import locale as Locale

from .global_state import global_state


def _exec(args):
    if not global_state.demo_mode:
        subprocess.run(args)


def _run_program(args):
    env = os.environ.copy()
    env["LANG"] = global_state.get_config('locale')
    Popen(args, env=env)


### public methods ###
def is_booted_with_uefi():
    return os.path.isdir("/sys/firmware/efi/efivars")


def open_disks():
    _run_program(global_state.get_config('disks_cmd').split())


def open_internet_search():
    browser_cmd = global_state.get_config('browser_cmd').split()
    failure_help_url = global_state.get_config('failure_help_url')
    version = global_state.get_config("version")
    browser_cmd.append(failure_help_url.format(version))
    _run_program(browser_cmd)


def open_wifi_settings():
    _run_program(global_state.get_config('wifi_cmd').split())


def reboot_system():
    _exec(['reboot'])


def set_system_keyboard_layout(keyboard_layout, short_hand):
    global_state.set_config('keyboard_layout_ui', keyboard_layout)
    global_state.set_config('keyboard_layout', short_hand)

    # set system input
    _exec(['gsettings', 'set', 'org.gnome.desktop.input-sources', 'sources',
           f"[('xkb','{short_hand}')]"])


def set_system_language(language_info):
    global_state.set_config('language', language_info.name)
    global_state.set_config('language_code', language_info.language_code)
    locale = Locale.normalize(language_info.locale)
    global_state.set_config('locale', locale)

    # set app language
    was_set = Locale.setlocale(Locale.LC_ALL, locale)
    if not was_set:
        print(f'Could not set locale to {language_info.name}, falling back to English.')
        print('Installation medium creators, check that you have correctly set up the locales',
              f'to support {language_info.name}.')
        # fallback
        Locale.setlocale(Locale.LC_ALL, 'en_US.UTF-8')

    # TODO find correct way to set system locale without user authentication
    _exec(['localectl', '--no-ask-password', 'set-locale', f'LANG={locale}'])


def set_system_formats(locale, formats_label):
    global_state.set_config('formats_locale', locale)
    global_state.set_config('formats_ui', formats_label)
    _exec(['gsettings', 'set', 'org.gnome.system.locale', 'region', f"'{locale}'"])


def set_system_timezone(timezone):
    global_state.set_config('timezone', timezone)
    # TODO find correct way to set timezone without user authentication
    _exec(['timedatectl', '--no-ask-password', 'set-timezone', timezone])


def start_system_timesync():
    # TODO find correct way to set enable time sync without user authentication
    _exec(['timedatectl', '--no-ask-password', 'set-ntp', 'true'])
    _exec(['gsettings', 'set', 'org.gnome.desktop.datetime', 'automatic-timezone', 'true'])



# SPDX-License-Identifier: GPL-3.0-or-later

import locale as Locale

from gi.repository import GnomeDesktop, GObject

from .global_state import global_state

locales = {
    'as_IN.UTF-8', 'pt_BR.UTF-8', 'am_ET.UTF-8', 'os_RU.UTF-8', 'ta_LK.UTF-8', 'ka_GE.UTF-8', 'ar_DZ.UTF-8',
    'nan_TW.UTF-8', 'el_GR.UTF-8', 'ta_IN.UTF-8', 'bo_CN.UTF-8', 'so_ET.UTF-8', 'sr_ME.UTF-8', 'hi_IN.UTF-8',
    'ru_RU.UTF-8', 'sk_SK.UTF-8', 'en_PH.UTF-8', 'en_BW.UTF-8', 'ur_PK.UTF-8', 'sm_WS.UTF-8', 'ik_CA.UTF-8',
    'sr_RS.UTF-8@latin', 'th_TH.UTF-8', 'en_US.UTF-8', 'tn_ZA.UTF-8', 'es_EC.UTF-8', 'zu_ZA.UTF-8',
    'nan_TW.UTF-8@latin', 'ar_SD.UTF-8', 'sw_KE.UTF-8', 'aa_ET.UTF-8', 'ar_TN.UTF-8', 'en_AG.UTF-8', 'es_ES.UTF-8',
    'aa_ER.UTF-8', 'gu_IN.UTF-8', 'pt_PT.UTF-8', 'he_IL.UTF-8', 'ar_AE.UTF-8', 'bg_BG.UTF-8', 'es_PE.UTF-8',
    'dv_MV.UTF-8', 'es_CR.UTF-8', 'kk_KZ.UTF-8', 'fi_FI.UTF-8', 'lg_UG.UTF-8', 'el_CY.UTF-8', 'fr_FR.UTF-8',
    'ar_EG.UTF-8', 'ak_GH.UTF-8', 'yo_NG.UTF-8', 'sr_RS.UTF-8', 'en_HK.UTF-8', 'es_MX.UTF-8', 'wae_CH.UTF-8',
    'lt_LT.UTF-8', 'fr_LU.UTF-8', 'kn_IN.UTF-8', 'te_IN.UTF-8', 'hak_TW.UTF-8', 've_ZA.UTF-8', 'dz_BT.UTF-8',
    'wa_BE.UTF-8', 'az_AZ.UTF-8', 'es_AR.UTF-8', 'gv_GB.UTF-8', 'hu_HU.UTF-8', 'ca_ES.UTF-8', 'mk_MK.UTF-8',
    'aa_ER.UTF-8@saaho', 'ar_SA.UTF-8', 'hy_AM.UTF-8', 'de_IT.UTF-8', 'lb_LU.UTF-8', 'mt_MT.UTF-8', 'en_IN.UTF-8',
    'sd_IN.UTF-8', 'ar_OM.UTF-8', 'bi_VU.UTF-8', 'kw_GB.UTF-8', 'en_NZ.UTF-8', 'sc_IT.UTF-8', 'li_NL.UTF-8',
    'tr_TR.UTF-8', 'af_ZA.UTF-8', 'uz_UZ.UTF-8@cyrillic', 'ts_ZA.UTF-8', 'so_DJ.UTF-8', 'en_IE.UTF-8', 'my_MM.UTF-8',
    'de_AT.UTF-8', 'ig_NG.UTF-8', 'ko_KR.UTF-8', 'de_BE.UTF-8', 'ms_MY.UTF-8', 'es_VE.UTF-8', 'sv_FI.UTF-8',
    'cy_GB.UTF-8', 'ar_IN.UTF-8', 'ha_NG.UTF-8', 'eu_ES.UTF-8', 'en_DK.UTF-8', 'yi_US.UTF-8', 'es_US.UTF-8',
    'mg_MG.UTF-8', 'ar_MA.UTF-8', 'cs_CZ.UTF-8', 'xh_ZA.UTF-8', 'nl_NL.UTF-8', 'es_DO.UTF-8', 'uk_UA.UTF-8',
    'mi_NZ.UTF-8', 'hr_HR.UTF-8', 'om_KE.UTF-8', 'eo.UTF-8', 'ja_JP.UTF-8', 'to_TO.UTF-8', 'be_BY.UTF-8@latin',
    'tt_RU.UTF-8@iqtelif', 'ar_IQ.UTF-8', 'it_IT.UTF-8', 'mn_MN.UTF-8', 'ug_CN.UTF-8', 'es_BO.UTF-8', 'rw_RW.UTF-8',
    'oc_FR.UTF-8', 'ss_ZA.UTF-8', 'ff_SN.UTF-8', 'ti_ER.UTF-8', 'sa_IN.UTF-8', 'es_CO.UTF-8', 'it_CH.UTF-8',
    'es_SV.UTF-8', 'vi_VN.UTF-8', 'fr_CA.UTF-8', 'de_LU.UTF-8', 'zh_CN.UTF-8', 'ro_RO.UTF-8', 'tk_TM.UTF-8',
    'ne_NP.UTF-8', 'nl_BE.UTF-8', 'ar_BH.UTF-8', 'es_CL.UTF-8', 'de_CH.UTF-8', 'ar_YE.UTF-8', 'sq_AL.UTF-8',
    'be_BY.UTF-8', 'is_IS.UTF-8', 'ar_QA.UTF-8', 'ky_KG.UTF-8', 'ca_ES.UTF-8@valencia', 'sv_SE.UTF-8', 'en_ZW.UTF-8',
    'lzh_TW.UTF-8', 'pa_PK.UTF-8', 'ar_JO.UTF-8', 'id_ID.UTF-8', 'ia_FR.UTF-8', 'de_LI.UTF-8', 'bn_BD.UTF-8',
    'es_HN.UTF-8', 'cv_RU.UTF-8', 'fy_NL.UTF-8', 'pa_IN.UTF-8', 'ks_IN.UTF-8@devanagari', 'ga_IE.UTF-8', 'nso_ZA.UTF-8',
    'sw_TZ.UTF-8', 'bo_IN.UTF-8', 'en_AU.UTF-8', 'wo_SN.UTF-8', 'lo_LA.UTF-8', 'az_IR.UTF-8', 'gl_ES.UTF-8',
    'ps_AF.UTF-8', 'gd_GB.UTF-8', 'sq_MK.UTF-8', 'no_NO.UTF-8', 'en_ZA.UTF-8', 'ar_SY.UTF-8', 'en_NG.UTF-8',
    'ru_UA.UTF-8', 'tg_TJ.UTF-8', 'da_DK.UTF-8', 'mr_IN.UTF-8', 'ln_CD.UTF-8', 'ti_ET.UTF-8', 'se_NO.UTF-8',
    'ce_RU.UTF-8', 'es_CU.UTF-8', 'li_BE.UTF-8', 'om_ET.UTF-8', 'an_ES.UTF-8', 'km_KH.UTF-8', 'nr_ZA.UTF-8',
    'nn_NO.UTF-8', 'ca_AD.UTF-8', 'zh_HK.UTF-8', 'lv_LV.UTF-8', 'es_PR.UTF-8', 'or_IN.UTF-8', 'fo_FO.UTF-8',
    'ca_IT.UTF-8', 'pl_PL.UTF-8', 'sd_IN.UTF-8@devanagari', 'en_IL.UTF-8', 'fr_CH.UTF-8', 'cmn_TW.UTF-8',
    'hif_FJ.UTF-8', 'ca_FR.UTF-8', 'en_GB.UTF-8', 'bn_IN.UTF-8', 'en_SG.UTF-8', 'nb_NO.UTF-8', 'es_PA.UTF-8',
    'tl_PH.UTF-8', 'ks_IN.UTF-8', 'fa_IR.UTF-8', 'tt_RU.UTF-8', 'si_LK.UTF-8', 'kl_GL.UTF-8', 'st_ZA.UTF-8',
    'bs_BA.UTF-8', 'zh_SG.UTF-8', 'br_FR.UTF-8', 'ur_IN.UTF-8', 'ar_SS.UTF-8', 'fr_BE.UTF-8', 'fy_DE.UTF-8',
    'es_UY.UTF-8', 'es_GT.UTF-8', 'es_NI.UTF-8', 'es_PY.UTF-8', 'et_EE.UTF-8', 'so_KE.UTF-8', 'en_ZM.UTF-8',
    'nl_AW.UTF-8', 'zh_TW.UTF-8', 'ht_HT.UTF-8', 'ar_LY.UTF-8', 'de_DE.UTF-8', 'ml_IN.UTF-8', 'sl_SI.UTF-8',
    'tr_CY.UTF-8', 'ar_KW.UTF-8', 'uz_UZ.UTF-8', 'ku_TR.UTF-8', 'en_CA.UTF-8', 'ar_LB.UTF-8', 'iu_CA.UTF-8',
    'aa_DJ.UTF-8', 'so_SO.UTF-8'}


class Format(GObject.GObject):
    __gtype_name__ = __qualname__
    name: str
    lower_case_name: str
    locale: str

    def __init__(self, name, locale):
        super().__init__()

        self.name = name
        self.lower_case_name = name.lower()
        self.locale = locale


### public methods ###

def get_current_formats():
    formats = global_state.get_config('formats_locale')
    if not formats:
        formats = global_state.get_config('locale')
        global_state.set_config('formats_locale', formats)
    name = GnomeDesktop.get_country_from_locale(formats)
    if not name:
        # solely to prevent crashes, e.g. for Esperanto
        # TODO add to translatation
        name = 'Undefined'
    return (formats, name)


def get_formats():
    formats = []
    # separate name set to prevent duplicates in list
    # see gnome-desktop issue https://gitlab.gnome.org/GNOME/gnome-shell/-/issues/3610
    names = set()
    translation_locale = global_state.get_config('locale')

    for locale in locales:
        name = GnomeDesktop.get_country_from_locale(locale, translation_locale)
        if name and not name in names:
            names.add(name)
            formats.append(Format(name, locale))

    # return sorted (considers umlauts and such)
    return sorted(
        formats,
        key=lambda t: Locale.strxfrm(t.name)
    )



# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import GLib, GObject

from .preloadable import Preloadable

class DeviceInfo(GObject.GObject):
    __gtype_name__ = __qualname__

    name: str = None
    size: int
    size_text: str
    device_path: str

    def __init__(self, name, size, size_text, device_path):
        super().__init__()

        if name:
            self.name = name.strip()
        self.size = size
        self.size_text = size_text
        self.device_path = device_path


class Disk(DeviceInfo):
    partitions: list = []
    efi_partition: str = ''

    def __init__(self, name, size, size_text, device_path, partitions=None):
        super().__init__(name, size, size_text, device_path)

        if partitions:
            self.partitions, self.efi_partition = partitions

class DiskProvider(Preloadable):
    udisks_client = None

    EFI_PARTITION_GUID = 'C12A7328-F81F-11D2-BA4B-00A0C93EC93B'
    EFI_PARTITON_FLAGS = None

    def __init__(self):
        Preloadable.__init__(self, self._init_client)

    def _init_client(self):
        # avoids initializing udisks client in demo mode
        import gi                           # noqa: E402
        gi.require_version('UDisks', '2.0') # noqa: E402
        from gi.repository import UDisks
        self.EFI_PARTITON_FLAGS = UDisks.PartitionTypeInfoFlags.SYSTEM.numerator
        self.udisks_client = UDisks.Client.new_sync()

    def _get_one_partition(self, partition, block):
        # partition info
        partition_info = DeviceInfo(
            name=block.props.id_label,
            size=block.props.size,
            size_text=self.disk_size_to_str(block.props.size),
            device_path=block.props.device)

        # check if EFI System Partiton
        is_efi_partition = (partition.props.type.upper() == self.EFI_PARTITION_GUID)

        # add to disk info
        return (partition_info, is_efi_partition)

    def _get_partitions(self, partition_table):
        if not partition_table:
            return None

        partitions = []
        efi_partition = ''
        for partition_name in partition_table.props.partitions:
            partition_object = self.udisks_client.get_object(partition_name)
            if not partition_object:
                continue
            block = partition_object.get_block()
            partition = partition_object.get_partition()
            if block and partition:
                partition_info, is_efi_partition = self._get_one_partition(partition, block)

                partitions.append(partition_info)
                if is_efi_partition:
                    efi_partition = partition_info.device_path
            else:
                print('Unhandled partiton in partition table, ignoring.')

        return (partitions, efi_partition)

    def _get_disk_info(self, block, drive, partition_table):
        # disk info
        disk = Disk(
            name=drive.props.vendor + ' ' + drive.props.model,
            size=block.props.size,
            size_text=self.disk_size_to_str(block.props.size),
            device_path=block.props.device,
            partitions=self._get_partitions(partition_table))

        return disk

    ### public methods ###

    def disk_size_to_str(self, size):
        self.assert_preloaded()
        return self.udisks_client.get_size_for_display(size, False, False)

    def get_disks(self):
        self.assert_preloaded()
        
        disks = []

        # get available devices
        dummy_var = GLib.Variant('a{sv}', None)
        devices = self.udisks_client.get_manager().call_get_block_devices_sync(dummy_var, None)

        # get device information
        for device in devices:
            udisks_object = self.udisks_client.get_object(device)
            if not udisks_object:
                continue

            # skip partitions
            partition = udisks_object.get_partition()
            if partition:
                continue

            block = udisks_object.get_block()
            if not block:
                continue

            partition_table = udisks_object.get_partition_table()
            drive = self.udisks_client.get_drive_for_block(block)
            if drive and not drive.props.optical:
                disk_info = self._get_disk_info(block, drive, partition_table)
                disks.append(disk_info)

        return disks

    def get_testing_dummy_disks(self):
        self.flip = not self.flip if hasattr(self, 'flip') and self.flip else True
        if not self.flip:
            return []
        smol_partition = DeviceInfo("sm0l partiton", 1000, "1 KB", "/dev/00null")
        smol_disk = Disk("Dummy", 10000, "10 KB", "/dev/null", ([smol_partition], None))

        efi_partition = DeviceInfo("EFI", 200000000, "20 GB", "/dev/sda_efi")
        unnamed_partition_1 = DeviceInfo(None, 20000000000, "20 GB", "/dev/sda_unnamed")
        unnamed_partition_2 = DeviceInfo(None, 20000000000, "20 GB", "/dev/sda_unnamed2")
        unnamed_partition_3 = DeviceInfo(None, 20000000000, "20 GB", "/dev/sda_unnamed3")
        partytion = DeviceInfo("PARTYtion", 20000000000, "20 GB", "/dev/sda_party")
        disk = Disk("Totally real device", 100000000000, "100 GB", "/dev/sda", ([efi_partition, partytion, unnamed_partition_1, unnamed_partition_2, unnamed_partition_3], "EFI"))

        unformated_big_disk = Disk("VERY BIG DISK", 1000000000000000, "1000 TB", "/dev/sdb_very_big")

        return [smol_disk, disk, unformated_big_disk]


disk_provider = DiskProvider()



# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Gtk

from .global_state import global_state
from .installation_scripting import installation_scripting, Step
from .page import Page
from .widgets import reset_model, DeviceRow


@Gtk.Template(resource_path='/com/github/p3732/os-installer/ui/pages/confirm.ui')
class ConfirmPage(Gtk.Box, Page):
    __gtype_name__ = __qualname__
    image = 'question-round-symbolic'

    disk_row = Gtk.Template.Child()

    def __init__(self, **kwargs):
        Gtk.Box.__init__(self, **kwargs)

    ### callbacks ###

    @Gtk.Template.Callback('confirmed')
    def _confirmed(self, button):
        installation_scripting.set_ok_to_start_step(Step.install)
        global_state.advance(self, allow_return=False)

    ### public methods ###

    def load(self):
        self.disk_row.set_title(global_state.get_config('disk_name'))
        self.disk_row.set_subtitle(global_state.get_config('disk_device_path'))



# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import GObject
from typing import NamedTuple

from .global_state import global_state
from .preloadable import Preloadable


class Option(NamedTuple):
    display : str
    keyword : str


class Choice(GObject.GObject):
    __gtype_name__ = __qualname__

    def __init__(self, name, description, icon_path, suggested=False, keyword=None, options=[]):
        super().__init__()

        self.name = name
        self.description = description
        self.icon_path = icon_path

        if options:
            self.options=options
        else:
            self.options=None
            self.keyword = keyword
            self.suggested = suggested


def handle_choice(choice):
    name = choice['name']
    description = choice['description'] if 'description' in choice else ''
    icon_path = choice['icon_path'] if 'icon_path' in choice else ''

    if 'options' in choice:
        if 'keyword' in choice or 'suggested' in choice:
            print(f"Config of {name}: 'options' can't be used with 'keyword'/'suggested'")
            return None

        options = []
        for option in choice['options']:
            if not 'option' in option:
                print(f'Option for {name} not correctly configured: {option}')
                continue
            option_name = option['name'] if 'name' in option else option['option']
            options.append(Option(option_name, option['option']))

        if len(options) == 0:
            print(f'No valid options found for {name}')
            return None
        else:
            return Choice(name, description, icon_path, options=options)
    else:
        if 'keyword' in choice:
            suggested = choice['suggested'] if 'suggested' in choice else False
            return Choice(name, description, icon_path, suggested=suggested, keyword=choice['keyword'])
        else:
            print(f'No keyword found for {name}')
            return None

def handle_legacy(choice):
    if 'package' in choice:
        print("Syntax changed! Use 'keyword' instead of 'package'")
        choice['keyword'] = choice['package']
    if 'feature' in choice:
        print("Syntax changed! Use 'keyword' instead of 'feature'")
        choice['keyword'] = choice['feature']


def handle_choices(config_entries):
    if not config_entries:
        return []

    choices: list = []
    for choice in config_entries:
        handle_legacy(choice)
        if (not 'name' in choice or not
                ('options' in choice or 'keyword' in choice)):
            print(f'Choice not correctly configured: {choice}')
            continue
        if parsed := handle_choice(choice):
            choices.append(parsed)

    return choices


class ChoicesProvider(Preloadable):
    def __init__(self):
        Preloadable.__init__(self, self._get_choices)

    def _get_choices(self):
        self.features = handle_choices(
            global_state.get_config('additional_features'))
        self.software = handle_choices(
            global_state.get_config('additional_software'))

    ### public methods ###

    def get_software_suggestions(self):
        self.assert_preloaded()
        return self.software

    def get_feature_suggestions(self):
        self.assert_preloaded()
        return self.features


choices_provider = ChoicesProvider()



# SPDX-License-Identifier: GPL-3.0-or-later

import time
from threading import Lock, Thread
from urllib.request import urlopen

from .global_state import global_state


def check_connection(url):
    try:
        urlopen(url, timeout=50)
        return True
    except:
        return False


class InternetProvider():
    callback_lock = Lock()
    callback = None

    def __init__(self):
        self.url = global_state.get_config('internet_checker_url')

        # connection locking
        self.connected = False

        # start internet connection checking
        self._start_connection_checker()
        self.thread = Thread(target=self._poll, daemon=True)
        self.thread.start()

    def _start_connection_checker(self):
        self.connection_checker = global_state.thread_pool.submit(check_connection, url=self.url)

    def _poll(self):
        connected = False
        while not connected:
            if not self.connection_checker.done():
                time.sleep(0.5)  # wait 500ms
            else:
                connected = self.connection_checker.result()

                if not connected:
                    # restart checker
                    self._start_connection_checker()

        with self.callback_lock:
            self.connected = connected
            if not self.callback is None:
                self.callback()

    ### public methods ###

    def is_connected_now_or_later(self, callback):
        ''' Returns `True` immidiately if connected. Otherwise returns False and calls callback when connected.'''
        with self.callback_lock:
            if self.connected:
                return True
            else:
                self.callback = callback
                return False


internet_provider = InternetProvider()



# SPDX-License-Identifier: GPL-3.0-or-later

from threading import Lock

from gi.repository import Gtk

from .global_state import global_state
from .internet_provider import internet_provider
from .page import Page
from .system_calls import open_wifi_settings, start_system_timesync


@Gtk.Template(resource_path='/com/github/p3732/os-installer/ui/pages/internet.ui')
class InternetPage(Gtk.Box, Page):
    __gtype_name__ = __qualname__
    image = 'network-wireless-disabled-symbolic'

    can_proceed_automatically = False
    connected = False
    connected_lock = Lock()

    def __init__(self, **kwargs):
        Gtk.Box.__init__(self, **kwargs)

    def _set_connected(self):
        self.image = 'network-wireless-symbolic'
        global_state.reload_title_image()
        start_system_timesync()

    ### callbacks ###

    @Gtk.Template.Callback('clicked_settings_button')
    def _clicked_settings_button(self, button):
        open_wifi_settings()

    def _on_connected(self):
        with self.connected_lock:
            self._set_connected()

        # do not hold lock, could cause deadlock with simultaneous load()
        global_state.advance(self)

    ### public methods ###

    def load_once(self):
        with self.connected_lock:
            # setup callback on connected
            if internet_provider.is_connected_now_or_later(self._on_connected):
                # already connected
                self._set_connected()
                return "load_next"
            if global_state.demo_mode:
                return "load_next"



# SPDX-License-Identifier: GPL-3.0-or-later

from threading import Thread

from .choices_provider import choices_provider
from .disk_provider import disk_provider
from .language_provider import language_provider


class PreloadManager:
    def __init__(self):
        self.thread = Thread(target=self._preload)

    def _preload(self):
        language_provider.preload()
        language_provider.assert_preloaded()

        disk_provider.preload()
        disk_provider.assert_preloaded()

        choices_provider.preload()
        choices_provider.assert_preloaded()

    ### public methods ###

    def start(self):
        self.thread.start()


preload_manager = PreloadManager()



# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Adw, Gtk


def reset_model(model, new_values):
    '''
    Reset given model to contain the passed new values.
    (Convenience wrapper)
    '''
    n_prev_items = model.get_n_items()
    model.splice(0, n_prev_items, new_values)

@Gtk.Template(resource_path='/com/github/p3732/os-installer/ui/widgets/device_row.ui')
class DeviceRow(Adw.ActionRow):
    __gtype_name__ = __qualname__

    stack = Gtk.Template.Child()
    size_label = Gtk.Template.Child()
    too_small_label = Gtk.Template.Child()

    def __init__(self, info, required_size_str=None, **kwargs):
        super().__init__(**kwargs)

        self.info = info
        self.size_label.set_label(info.size_text)
        if info.name:
            self.set_title(info.name)

        self.set_subtitle(info.device_path)

        if required_size_str:
            smol = self.too_small_label.get_label()
            self.too_small_label.set_label(smol.format(required_size_str))
            self.set_activatable(False)
            self.set_sensitive(False)
            self.stack.set_visible_child_name('too_small')


@Gtk.Template(resource_path='/com/github/p3732/os-installer/ui/widgets/multi_selection_row.ui')
class MultiSelectionRow(Adw.ComboRow):
    __gtype_name__ = __qualname__

    icon = Gtk.Template.Child()
    list = Gtk.Template.Child()

    def __init__(self, title, description, icon_path, fallback_icon,
                 options, **kwargs):
        super().__init__(**kwargs)
        self.set_title(title)
        self.set_subtitle(description)
        if not icon_path:
            self.icon.set_from_icon_name(fallback_icon)
            self.icon.set_icon_size(Gtk.IconSize.LARGE)
        else:
            self.icon.set_from_file(icon_path)

        self.icon_path = icon_path
        self.options = options

        self.list.splice(0, 0, [option.display for option in options])
        self.set_model(self.list)

    def get_chosen_option(self):
        return self.options[self.get_selected()]


@Gtk.Template(resource_path='/com/github/p3732/os-installer/ui/widgets/page_wrapper.ui')
class PageWrapper(Gtk.Box):
    __gtype_name__ = __qualname__

    content = Gtk.Template.Child()

    def __init__(self, page, **kwargs):
        super().__init__(**kwargs)

        self.content.set_child(page)

    def get_page(self):
        return self.content.get_child()

@Gtk.Template(resource_path='/com/github/p3732/os-installer/ui/widgets/progress_row.ui')
class ProgressRow(Gtk.ListBoxRow):
    __gtype_name__ = __qualname__

    title = Gtk.Template.Child()

    def __init__(self, label, additional_info = None, **kwargs):
        super().__init__(**kwargs)

        self.title.set_label(label)

        self.info = additional_info

    def get_label(self):
        return self.title.get_label()


@Gtk.Template(resource_path='/com/github/p3732/os-installer/ui/widgets/selection_row.ui')
class SelectionRow(Adw.ActionRow):
    __gtype_name__ = __qualname__

    icon = Gtk.Template.Child()
    switch = Gtk.Template.Child()

    def __init__(self, title, description, icon_path, fallback_icon,
                 default_state, info, **kwargs):
        super().__init__(**kwargs)
        self.set_title(title)
        self.set_subtitle(description)
        self.switch.set_active(default_state)
        if not icon_path:
            self.icon.set_from_icon_name(fallback_icon)
            self.icon.set_icon_size(Gtk.IconSize.LARGE)
        else:
            self.icon.set_from_file(icon_path)

        self.icon_path = icon_path
        self.info = info

    def is_activated(self):
        return self.switch.get_active()

    def flip_switch(self):
        return self.switch.set_active(not self.switch.get_active())


@Gtk.Template(resource_path='/com/github/p3732/os-installer/ui/widgets/summary_row.ui')
class SummaryRow(Gtk.ListBoxRow):
    __gtype_name__ = __qualname__

    icon = Gtk.Template.Child()
    name = Gtk.Template.Child()

    def __init__(self, name, icon_path, icon_name, **kwargs):
        super().__init__(**kwargs)
        self.name.set_label(name)
        if not icon_path:
            self.icon.set_from_icon_name(icon_name)
        else:
            self.icon.set_from_file(icon_path)



# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Gtk

from .global_state import global_state
from .language_provider import language_provider
from .page import Page
from .system_calls import set_system_language
from .widgets import reset_model, ProgressRow


@Gtk.Template(resource_path='/com/github/p3732/os-installer/ui/pages/language.ui')
class LanguagePage(Gtk.Box, Page):
    __gtype_name__ = __qualname__
    image = 'language-symbolic'

    suggested_list = Gtk.Template.Child()
    suggested_model = Gtk.Template.Child()

    other_list = Gtk.Template.Child()
    other_model = Gtk.Template.Child()

    language_chosen = False

    def __init__(self, **kwargs):
        Gtk.Box.__init__(self, **kwargs)

        # models
        self.suggested_list.bind_model(self.suggested_model, lambda o: ProgressRow(o.name, o))
        self.other_list.bind_model(self.other_model, lambda o: ProgressRow(o.name, o))

    ### callbacks ###

    @Gtk.Template.Callback('language_row_activated')
    def _language_row_activated(self, list_box, row):
        if (not self.language_chosen or
                global_state.get_config('language_code') != row.info.language_code):
            self.language_chosen = True
            set_system_language(row.info)
            global_state.load_translated_pages()
        global_state.advance(self)

    ### public methods ###

    def load_once(self):
        suggested_languages = language_provider.get_suggested_languages()
        other_languages = language_provider.get_other_languages()
        if not suggested_languages:
            self.suggested_list.set_visible(False)
        if not other_languages:
            self.other_list.set_visible(False)
        
        reset_model(self.suggested_model, suggested_languages)
        reset_model(self.other_model, other_languages)



# SPDX-License-Identifier: GPL-3.0-or-later

from time import time
from gi.repository import GnomeDesktop, GObject, GWeather


class Timezone(GObject.GObject):
    __gtype_name__ = __qualname__

    def __init__(self, name):
        super().__init__()

        self.name: str = name
        self.lower_case_name: str = name.lower()
        self.locations: set = set()


def _get_location_children(location):
    # this code is un-pythonesque because libgweather decided to simplify their API too much
    first_child = location.next_child(None)
    if not first_child:
        return []
    children = [first_child]
    while child := location.next_child(children[-1]):
        children.append(child)
    return children


def _add_all_locations_to_timezone(timezone, location):
    for child in _get_location_children(location):
        timezone.locations.add(child.get_name().lower())
        _add_all_locations_to_timezone(timezone, child)


def _recurse_location(location, timezones):
    for child in _get_location_children(location):
        if child.has_timezone():
            timezone_id = child.get_timezone().get_identifier()
            if not timezone_id in timezones:
                print(f'Developer hint: Unknown timezone {timezone_id} {child.get_name()}')
                continue
            _add_all_locations_to_timezone(timezones[timezone_id], child)
        else:
            _recurse_location(child, timezones)


### public methods ###

def get_current_timezone():
    timezone = GnomeDesktop.WallClock().get_timezone()
    return timezone.get_identifier()


def get_timezones():
    timezones = {timezone.get_identifier(): Timezone(timezone.get_identifier())
                 for timezone in GWeather.Location().get_world().get_timezones()}
    for child in _get_location_children(GWeather.Location().get_world()):
        if not child.has_timezone(): # skips UTC and Etc/GMT+12
            _recurse_location(child, timezones)

    return sorted(timezones.values(), key=lambda t: t.name)



# SPDX-License-Identifier: GPL-3.0-or-later

import yaml

DEFAULT_CONFIG_PATH = '/etc/os-installer/config.yaml'


def _bool_to_int(b: bool):
    return 1 if b else 0


def _match(config, prop, *ok_types):
    if not prop in config:
        print(f'Config error: {prop} does not exist.')
        return False
    has_type = type(config[prop])
    for ok_type in ok_types:
        if has_type == ok_type:
            return True
    print(f'Config error: {prop} not of expected type (expected',
          f'{ok_types}, but got {has_type}')
    return False


def _install_variables_set(config):
    install_variables = ['locale', 'disk_device_path', 'disk_is_partition',
                         'disk_efi_partition', 'use_encryption', 'encryption_pin']
    return all([var in config for var in install_variables])


def _configure_variables_set(config):
    configure_variables = ['user_name', 'user_autologin', 'user_password',
                           'formats_locale', 'timezone', 'chosen_software_packages', 
                           'chosen_feature_names']
    return all([var in config for var in configure_variables])


def _load_default_config():
    return {
        # general
        'distribution_name': 'Untitled',
        # internet
        'internet_connection_required': True,
        'internet_checker_url': 'http://nmcheck.gnome.org/check_network_status.txt',
        # language
        'suggested_languages': ['en', 'ar', 'de', 'es', 'fr', 'ja', 'ru', 'zh'],
        'fixed_language': False,
        # welcome
        'welcome_page': {'usage': True, 'logo': None, 'text': None},
        # keyboard
        'keyboard_layout' : 'us',
        # disk
        'minimum_disk_size': 5,
        'offer_disk_encryption': True,
        # optional pages
        'skip_user': False,
        'skip_locale': False,
        # software
        'additional_software': [],
        # feature
        'additional_features': [],
        # fail
        'failure_help_url': 'https://duckduckgo.com/?q="os-installer {}"+"failed installation"',
        # commands
        'browser_cmd': 'epiphany',
        'disks_cmd': 'gnome-disks',
        'wifi_cmd': 'gnome-control-center wifi',

        # not configurable
        'version': -1,
    }


def _load_optional_defaults(config):
    config['use_encryption'] = False
    config['encryption_pin'] = ''
    config['user_name'] = 'user'
    config['user_autologin'] = True
    config['user_password'] = 'password'
    config['formats_locale'] = 'en_US.UTF-8'
    config['formats_ui'] = 'United States'
    config['timezone'] = 'UTC'
    config['chosen_software_packages'] = ''
    config['chosen_software'] = []
    config['chosen_feature_names'] = ''
    config['chosen_features'] = []


def _set_testing_defaults(config):
    '''Default values used when skipping pages during testing.'''
    config['language'] = 'English for Dummies'
    config['language_code'] = 'en_US'
    config['locale'] = 'en_US.UTF-8'
    config['keyboard_layout'] = 'us'
    config['keyboard_layout_ui'] = 'English (US)'
    config['disk_device_path'] = '/dev/null'
    config['disk_is_partition'] = False
    config['disk_efi_partition'] = '/dev/null'
    config['disk_name'] = 'Test Dummy'


def _valid(config):
    assert not config['fixed_language'] == True, 'Need to specify or disable fixed language.'
    return (
        _match(config, 'welcome_page', dict) and
        _match(config['welcome_page'], 'usage', bool) and
        _match(config['welcome_page'], 'logo', str, type(None)) and
        _match(config['welcome_page'], 'text', str, type(None)) and
        _match(config, 'internet_connection_required', bool) and
        _match(config, 'internet_checker_url', str) and
        _match(config, 'suggested_languages', list) and
        _match(config, 'minimum_disk_size', int) and
        _match(config, 'offer_disk_encryption', bool) and
        _match(config, 'additional_software', list) and
        _match(config, 'additional_features', list) and
        _match(config, 'distribution_name', str) and
        _match(config, 'fixed_language', bool, str))


### public methods ###
def init_config():
    config = _load_default_config()
    try:
        with open(DEFAULT_CONFIG_PATH, 'r') as file:
            config_from_file = yaml.load(file, Loader=yaml.Loader)
            for config_property in config_from_file:
                if type(config[config_property]) is dict:
                    for key, value in config_from_file[config_property].items():
                        config[config_property][key] = value
                else:
                    config[config_property] = config_from_file[config_property]
    except:
        print('No config provided or config contains syntax errors, '
              'using default config.')
        config = _load_default_config()
    if not _valid(config):
        print('Config errors, loading default config.')
        config = _load_default_config()
    _load_optional_defaults(config)
    _set_testing_defaults(config)
    return config


def create_envs(config, with_install_envs, with_configure_envs):
    if (with_install_envs and not _install_variables_set(config) or
            with_configure_envs and not _configure_variables_set(config)):
        return None
    envs = []
    if with_install_envs:
        envs += [
            f'OSI_LOCALE={config["locale"]}',
            f'OSI_KEYBOARD_LAYOUT={config["keyboard_layout"]}',
            f'OSI_DEVICE_PATH={config["disk_device_path"]}',
            f'OSI_DEVICE_IS_PARTITION={_bool_to_int(config["disk_is_partition"])}',
            f'OSI_DEVICE_EFI_PARTITION={config["disk_efi_partition"]}',
            f'OSI_USE_ENCRYPTION={_bool_to_int(config["use_encryption"])}',
            f'OSI_ENCRYPTION_PIN={config["encryption_pin"]}',
        ]
    if with_configure_envs:
        envs += [
            f'OSI_USER_NAME={config["user_name"]}',
            f'OSI_USER_AUTOLOGIN={_bool_to_int(config["user_autologin"])}',
            f'OSI_USER_PASSWORD={config["user_password"]}',
            f'OSI_FORMATS={config["formats_locale"]}',
            f'OSI_TIMEZONE={config["timezone"]}',
            f'OSI_ADDITIONAL_SOFTWARE={config["chosen_software_packages"]}',
            f'OSI_ADDITIONAL_FEATURES={config["chosen_feature_names"]}',
        ]
    return envs + [None]



# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Gtk

from .format_provider import get_current_formats
from .global_state import global_state
from .timezone_provider import get_current_timezone
from .system_calls import set_system_formats
from .page import Page


@Gtk.Template(resource_path='/com/github/p3732/os-installer/ui/pages/locale.ui')
class LocalePage(Gtk.Box, Page):
    __gtype_name__ = __qualname__
    image = 'globe-symbolic'

    formats_label = Gtk.Template.Child()
    timezone_label = Gtk.Template.Child()

    def __init__(self, **kwargs):
        Gtk.Box.__init__(self, **kwargs)

    ### callbacks ###

    @Gtk.Template.Callback('continue')
    def _continue(self, button):
        global_state.advance(self)

    @Gtk.Template.Callback('overview_row_activated')
    def _overview_row_activated(self, list_box, row):
        global_state.navigate_to_page(row.get_name())

    ### public methods ###

    def load(self):
        formats = global_state.get_config('formats_ui')
        if not formats:
            locale, name = get_current_formats()
            set_system_formats(locale, name)
            formats = global_state.set_config('formats_ui', name)
        self.formats_label.set_label(formats)

        timezone = global_state.get_config('timezone')
        if not timezone:
            timezone = get_current_timezone()
            global_state.set_config('timezone', timezone)
        self.timezone_label.set_label(timezone)



# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import GObject


class SummaryEntry(GObject.GObject):
    __gtype_name__ = __qualname__

    def __init__(self, name, icon_path):
        super().__init__()

        self.name = name
        self.icon_path = icon_path



# SPDX-License-Identifier: GPL-3.0-or-later

from  pathlib import Path
from  typing import Union

class Page:
    image: Union[str, Path, None] = None
    can_navigate_backward: bool = False
    can_navigate_forward: bool = False
    can_reload: bool = False

    loaded: bool = False

    def id(self):
        return self.__gtype_name__

    ### dummy stubs ###

    def navigate_backward(self):
        '''
        Called upon backward navigation if `can_navigate_backward` is True.
        '''
        return

    def navigate_forward(self):
        '''
        Called upon forward navigation if `can_navigate_forward` is True.
        '''
        return

    def load(self):
        '''
        Called before the page is shown.
        Pages can overwrite this to receive call every time.
        Returning True means the page can be skipped.
        '''
        if not self.loaded:
            self.loaded = True
            return self.load_once()

    def load_once(self):
        '''
        Called once on first page construction. Used for e.g. filling lists.
        Special return values: "load_next" (skips page), "prevent_back_navigation"
        '''
        return

    def unload(self):
        '''
        Called before the page is no longer shown. Used for e.g. storing current enrty values.
        '''
        return



# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Gtk

from .global_state import global_state
from .installation_scripting import installation_scripting
from .system_calls import open_internet_search
from .page import Page


@Gtk.Template(resource_path='/com/github/p3732/os-installer/ui/pages/failed.ui')
class FailedPage(Gtk.Box, Page):
    __gtype_name__ = __qualname__
    image = 'computer-fail-symbolic'

    page_title = Gtk.Template.Child()
    terminal_box = Gtk.Template.Child()

    def __init__(self, **kwargs):
        Gtk.Box.__init__(self, **kwargs)

    ### callbacks ###

    @Gtk.Template.Callback('search_button_clicked')
    def _search_button_clicked(self, button):
        open_internet_search()

    ### public methods ###

    def load_once(self):
        self.terminal_box.append(installation_scripting.terminal)
        global_state.send_notification(self.page_title.get_label(),'')
        return 'load_next'



# SPDX-License-Identifier: GPL-3.0-or-later

from threading import Lock
import os

from gi.repository import GObject, GnomeDesktop

from .global_state import global_state
from .preloadable import Preloadable


# generated via language_codes_to_x_generator.py
language_to_default_locale = {
    'ab': 'ab_GE.UTF-8', 'aa': 'aa_DJ.UTF-8', 'af': 'af_ZA.UTF-8', 'ak': 'ak_GH.UTF-8', 'sq': 'sq_AL.UTF-8',
    'am': 'am_ET.UTF-8', 'ar': 'ar_EG.UTF-8', 'an': 'an_ES.UTF-8', 'hy': 'hy_AM.UTF-8', 'as': 'as_IN.UTF-8',
    'ae_AE': 'ar_AE.UTF-8', 'az': 'az_AZ.UTF-8', 'ba_BA': 'bs_BA.UTF-8', 'eu': 'eu_ES.UTF-8', 'be': 'be_BY.UTF-8',
    'bn': 'bn_BD.UTF-8', 'bh_BH': 'ar_BH.UTF-8', 'bi': 'bi_VU.UTF-8', 'bs': 'bs_BA.UTF-8', 'br': 'br_FR.UTF-8',
    'bg': 'bg_BG.UTF-8', 'my': 'my_MM.UTF-8', 'ca': 'ca_ES.UTF-8', 'ch_CH': 'de_CH.UTF-8', 'ce': 'ce_RU.UTF-8',
    'zh': 'zh_CN.UTF-8', 'cv': 'cv_RU.UTF-8', 'kw': 'kw_GB.UTF-8', 'co_CO': 'es_CO.UTF-8', 'cr_CR': 'es_CR.UTF-8',
    'hr': 'hr_HR.UTF-8', 'cs': 'cs_CZ.UTF-8', 'da': 'da_DK.UTF-8', 'dv': 'dv_MV.UTF-8', 'nl': 'nl_NL.UTF-8', 'dz':
    'dz_BT.UTF-8', 'en': 'en_US.UTF-8', 'en_GB': 'en_GB.UTF-8', 'eo': 'eo.UTF-8', 'et': 'et_EE.UTF-8',
    'ee_EE': 'et_EE.UTF-8', 'fo': 'fo_FO.UTF-8', 'fj_FJ': 'hif_FJ.UTF-8', 'fi': 'fi_FI.UTF-8', 'fr': 'fr_FR.UTF-8',
    'ff': 'ff_SN.UTF-8', 'gl': 'gl_ES.UTF-8', 'ka': 'ka_GE.UTF-8', 'de': 'de_DE.UTF-8', 'el': 'el_GR.UTF-8',
    'gu': 'gu_IN.UTF-8', 'ht': 'ht_HT.UTF-8', 'ha': 'ha_NG.UTF-8', 'he': 'he_IL.UTF-8', 'hi': 'hi_IN.UTF-8',
    'hu': 'hu_HU.UTF-8', 'ia': 'ia_FR.UTF-8', 'id': 'id_ID.UTF-8', 'ie_IE': 'en_IE.UTF-8', 'ga': 'ga_IE.UTF-8',
    'ig': 'ig_NG.UTF-8', 'ik': 'ik_CA.UTF-8', 'is': 'is_IS.UTF-8', 'it': 'it_IT.UTF-8', 'iu': 'iu_CA.UTF-8',
    'ja': 'ja_JP.UTF-8', 'kl': 'kl_GL.UTF-8', 'kn': 'kn_IN.UTF-8', 'kr_KR': 'ko_KR.UTF-8', 'ks': 'ks_IN.UTF-8@devanagari',
    'kk': 'kk_KZ.UTF-8', 'km': 'km_KH.UTF-8', 'rw': 'rw_RW.UTF-8', 'ky': 'ky_KG.UTF-8', 'kg_KG': 'ky_KG.UTF-8',
    'ko': 'ko_KR.UTF-8', 'ku': 'ku_TR.UTF-8', 'la_LA': 'lo_LA.UTF-8', 'lb': 'lb_LU.UTF-8', 'lg': 'lg_UG.UTF-8',
    'li': 'li_NL.UTF-8', 'ln': 'ln_CD.UTF-8', 'lo': 'lo_LA.UTF-8', 'lt': 'lt_LT.UTF-8', 'lu_LU': 'fr_LU.UTF-8',
    'lv': 'lv_LV.UTF-8', 'gv': 'gv_GB.UTF-8', 'mk': 'mk_MK.UTF-8', 'mg': 'mg_MG.UTF-8', 'ms': 'ms_MY.UTF-8',
    'ml': 'ml_IN.UTF-8', 'mt': 'mt_MT.UTF-8', 'mi': 'mi_NZ.UTF-8', 'mr': 'mr_IN.UTF-8', 'mn': 'mn_MN.UTF-8',
    'ne': 'ne_NP.UTF-8', 'ng_NG': 'en_NG.UTF-8', 'nb': 'nb_NO.UTF-8', 'nn': 'nn_NO.UTF-8', 'no': 'no_NO.UTF-8',
    'nr': 'nr_ZA.UTF-8', 'oc': 'oc_FR.UTF-8', 'cu_CU': 'es_CU.UTF-8', 'om': 'om_ET.UTF-8', 'or': 'or_IN.UTF-8',
    'os': 'os_RU.UTF-8', 'pa': 'pa_IN.UTF-8', 'fa': 'fa_IR.UTF-8', 'pl': 'pl_PL.UTF-8', 'ps': 'ps_AF.UTF-8',
    'pt': 'pt_BR.UTF-8', 'ro': 'ro_RO.UTF-8', 'ru': 'ru_RU.UTF-8', 'sa': 'sa_IN.UTF-8', 'sc': 'sc_IT.UTF-8',
    'sd': 'sd_IN.UTF-8', 'se': 'se_NO.UTF-8', 'sm': 'sm_WS.UTF-8', 'sg_SG': 'en_SG.UTF-8', 'sr': 'sr_RS.UTF-8',
    'gd': 'gd_GB.UTF-8', 'sn_SN': 'wo_SN.UTF-8', 'si': 'si_LK.UTF-8', 'sk': 'sk_SK.UTF-8', 'sl': 'sl_SI.UTF-8',
    'so': 'so_SO.UTF-8', 'st': 'st_ZA.UTF-8', 'es': 'es_ES.UTF-8', 'sw': 'sw_KE.UTF-8', 'ss': 'ss_ZA.UTF-8',
    'sv': 'sv_SE.UTF-8', 'ta': 'ta_IN.UTF-8', 'te': 'te_IN.UTF-8', 'tg': 'tg_TJ.UTF-8', 'th': 'th_TH.UTF-8',
    'ti': 'ti_ER.UTF-8', 'bo': 'bo_CN.UTF-8', 'tk': 'tk_TM.UTF-8', 'tl': 'tl_PH.UTF-8', 'tn': 'tn_ZA.UTF-8',
    'to': 'to_TO.UTF-8', 'tr': 'tr_TR.UTF-8', 'ts': 'ts_ZA.UTF-8', 'tt': 'tt_RU.UTF-8', 'tw_TW': 'zh_TW.UTF-8',
    'ug': 'ug_CN.UTF-8', 'uk': 'uk_UA.UTF-8', 'ur': 'ur_PK.UTF-8', 'uz': 'uz_UZ.UTF-8@cyrillic', 've': 've_ZA.UTF-8',
    'vi': 'vi_VN.UTF-8', 'wa': 'wa_BE.UTF-8', 'cy': 'cy_GB.UTF-8', 'wo': 'wo_SN.UTF-8', 'fy': 'fy_NL.UTF-8',
    'xh': 'xh_ZA.UTF-8', 'yi': 'yi_US.UTF-8', 'yo': 'yo_NG.UTF-8', 'za_ZA': 'zu_ZA.UTF-8', 'zu': 'zu_ZA.UTF-8',
    # manually added
    'pt_BR': 'pt_BR.UTF-8', 'pt_PT': 'pt_PT.UTF-8',
    }


class LanguageInfo(GObject.GObject):
    __gtype_name__ = __qualname__
    name: str
    language_code: str
    locale: str

    def __init__(self, name, language_code, locale):
        super().__init__()

        self.name = name
        self.language_code = language_code
        self.locale = locale


class LanguageProvider(Preloadable) :
    def __init__(self):
        Preloadable.__init__(self, self._get_languages)

    def _get_default_locale(self, language):
        if language in language_to_default_locale:
            return language_to_default_locale[language]
        else:
            return None

    def _get_existing_translations(self, localedir):
        # English always exists
        existing_translations = {'en'}

        # check what translations exist in the locale folder
        for file in os.scandir(localedir):
            if file.is_dir():
                locale_folder = os.path.join(file.path, 'LC_MESSAGES')
                if os.path.isdir(locale_folder):
                    for locale_file in os.scandir(locale_folder):
                        if locale_file.name == 'os-installer.mo':
                            language = os.path.basename(file.path)
                            existing_translations.add(language)
        return existing_translations

    def _get_language_name_localized(self, lang_locale, localization, lang_code):
        if name := GnomeDesktop.get_language_from_locale(lang_locale, localization):
            return name
        elif lang := GnomeDesktop.get_language_from_code(lang_code.split('_')[0], localization):
            return f'{lang} ({lang_code})'

    def _get_languages(self):
        localedir = global_state.get_config('localedir')
        translations = self._get_existing_translations(localedir)

        self.all_languages = []
        unavailable_languages = []
        for language_code in translations:
            locale = self._get_default_locale(language_code)

            if name := self._get_language_name_localized(locale, locale, language_code):
                language_info =  LanguageInfo(name, language_code, locale)
                self.all_languages.append(language_info)
            else:
                unavailable_languages.append(language_code)

        if unavailable_languages:
            print('The following locales are not available on the current system: ',
                  sorted(unavailable_languages))
        self.all_languages.sort(key=lambda k: k.name)

        self.suggested = []
        self.other = []
        suggested_codes = global_state.get_config('suggested_languages')
        if suggested_codes and len(suggested_codes) > 0:
            for language_info in self.all_languages:
                if language_info.language_code in suggested_codes:
                    self.suggested.append(language_info)
                else:
                    self.other.append(language_info)

    ### public methods ###

    def get_all_languages(self):
        self.assert_preloaded()
        return self.all_languages

    def get_fixed_language(self, fixed_language):
        self.assert_preloaded()

        fixed = [info for info in self.all_languages
                 if info.language_code == fixed_language]
        if len(fixed) == 1:
            return fixed[0]
        else:
            print('Distribution developer hint: Fixed language '
                  f'{fixed_language} is not available in current system. '
                  'Falling back to default language selection.')
            return None

    def get_suggested_languages(self):
        self.assert_preloaded()
        return self.suggested

    def get_other_languages(self):
        self.assert_preloaded()
        return self.other

language_provider = LanguageProvider()



# SPDX-License-Identifier: GPL-3.0-or-later

from pathlib import Path
from threading import Lock
from os.path import exists

from gi.repository import Gtk, Adw

from .global_state import global_state

from .choices import FeaturePage, SoftwarePage
from .confirm import ConfirmPage
from .disk import DiskPage
from .done import DonePage
from .encrypt import EncryptPage
from .failed import FailedPage
from .filter import FormatPage, TimezonePage
from .install import InstallPage
from .internet import InternetPage
from .keyboard_layout import KeyboardLayoutPage
from .language import LanguagePage
from .locale import LocalePage
from .restart import RestartPage
from .summary import SummaryPage
from .user import UserPage
from .welcome import WelcomePage
from .widgets import PageWrapper

from .language_provider import language_provider
from .system_calls import set_system_language


class Navigation:
    current: int = -1
    earliest: int = 0
    furthest: int = 0

    def set(self, state: int):
        self.current = state
        self.furthest = max(self.furthest, state)

    def is_not_earliest(self):
        return self.current > self.earliest

    def is_not_furthest(self):
        return self.current < self.furthest


@Gtk.Template(resource_path='/com/github/p3732/os-installer/ui/main_window.ui')
class OsInstallerWindow(Adw.ApplicationWindow):
    __gtype_name__ = __qualname__

    image_stack = Gtk.Template.Child()
    main_stack = Gtk.Template.Child()

    next_revealer = Gtk.Template.Child()
    previous_revealer = Gtk.Template.Child()
    reload_revealer = Gtk.Template.Child()

    current_page = None
    # when changing pages by name return to this page on advancing
    original_page_name: str = ''
    navigation_lock = Lock()
    navigation = Navigation()
    pages = []

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # set advancing functions in global state
        global_state.advance = self.advance
        global_state.load_translated_pages = self.load_translated_pages
        global_state.navigate_to_page = self.navigate_to_page
        global_state.reload_page = self.reload_page
        global_state.reload_title_image = self._reload_title_image
        global_state.installation_failed = self.show_failed_page

        # determine available pages
        self._determine_available_pages()

        if 'language' in self.available_pages:
            # only initialize language page, others depend on chosen language
            self._initialize_page('language')
        else:
            for page_name in self.available_pages.keys():
                self._initialize_page(page_name)

    def _determine_available_pages(self):
        # list page types tupled with condition on when to use
        pages = [
            # pre-installation section
            ('language', LanguagePage, self._offer_language_selection()),
            ('welcome', WelcomePage, global_state.get_config('welcome_page')['usage']),
            ('keyboard', KeyboardLayoutPage, True),
            ('internet', InternetPage, global_state.get_config(
                'internet_connection_required')),
            ('disk', DiskPage, True),
            ('encrypt', EncryptPage, global_state.get_config('offer_disk_encryption')),
            ('confirm', ConfirmPage, exists('/etc/os-installer/scripts/install.sh')),
            # configuration section
            ('user', UserPage, not global_state.get_config('skip_user')),
            ('locale', LocalePage, not global_state.get_config('skip_locale')),
            ('format', FormatPage, not global_state.get_config('skip_locale')),
            ('timezone', TimezonePage, not global_state.get_config('skip_locale')),
            ('software', SoftwarePage, global_state.get_config('additional_software')),
            ('feature', FeaturePage, global_state.get_config('additional_features')),
            # summary
            ('summary', SummaryPage, True),
            # installation
            ('install', InstallPage, True),
            # post-installation
            ('done', DonePage, True),
            ('restart', RestartPage, True),
            # failed installation
            ('failed', FailedPage, True)
        ]
        # filter out nonexistent pages
        self.available_pages = {page_name: page for page_name, page, condition in pages if condition}

    def _offer_language_selection(self):
            # only initialize language page, others depend on chosen language
        if fixed_language := global_state.get_config('fixed_language'):
            if fixed_info := language_provider.get_fixed_language(fixed_language):
                set_system_language(fixed_info)
                return False
            else:
                print('Developer hint: defined fixed language not available')
                global_state.set_config('fixed_language', '')
        return True

    def _initialize_page(self, page_name):
        page_type = self.available_pages[page_name]
        page = PageWrapper(page_type())

        self.main_stack.add_named(page, page_name)
        self.pages.append(page_name)

    def _remove_pages(self, page_names):
        for page_name in page_names:
            if page_name is not None: # skip already deleted pages
                child = self.main_stack.get_child_by_name(page_name)
                self.main_stack.remove(child)
                del child

    def _load_page(self, page_number: int):
        assert page_number >= 0, 'Tried to go to non-existent page (underflow)'
        assert page_number < len(self.pages), 'Tried to go to non-existent page (overflow)'

        # unload old page
        if self.current_page:
            self.current_page.unload()

        # load page
        page_name = self.pages[page_number]
        wrapper = self.main_stack.get_child_by_name(page_name)
        self.current_page = wrapper.get_page()

        match self.current_page.load():
            case "load_next":
                self._load_page(page_number + 1)
                return
            case "prevent_back_navigation":
                self.navigation.earliest = page_number

        self.main_stack.set_visible_child(wrapper)
        self.navigation.set(page_number)
        self._reload_title_image()
        self._update_navigation_buttons()

    def _load_page_by_name(self, page_name: str) -> None:
        self.current_page.unload()

        wrapper = self.main_stack.get_child_by_name(page_name)
        if wrapper == None:
            print(f'Page named {page_name} does not exist. Are you testing things and forget to comment it back in?')
            return
        self.current_page = wrapper.get_page()
        self.current_page.load()
        self.main_stack.set_visible_child(wrapper)

        self._reload_title_image()
        self.previous_revealer.set_reveal_child(True)
        self.next_revealer.set_reveal_child(False)
        self.reload_revealer.set_reveal_child(self.current_page.can_reload)

    def _load_original_page(self):
        self.current_page.unload()

        original_page = self.main_stack.get_child_by_name(self.original_page_name)
        self.current_page = original_page.get_page()
        self.current_page.load()
        self.main_stack.set_visible_child(original_page)
        self.original_page_name = None

        self._reload_title_image()
        self._update_navigation_buttons()

    def _reload_title_image(self):
        next_image_name = '1' if self.image_stack.get_visible_child_name() == '2' else '2'
        next_image = self.image_stack.get_child_by_name(next_image_name)
        image_source = self.current_page.image
        if isinstance(image_source, str):
            next_image.set_from_icon_name(image_source)
        elif isinstance(image_source, Path):
            next_image.set_from_file(str(image_source))
        else:
            print('Developer hint: invalid request to set title image')
            return # ignoring
        self.image_stack.set_visible_child_name(next_image_name)

    def _show_dialog(self, dialog):
        dialog.set_transient_for(self)
        dialog.set_modal(True)
        dialog.present()

    def _update_navigation_buttons(self):
        # backward
        show_backward = self.current_page.can_navigate_backward or self.navigation.is_not_earliest()
        self.previous_revealer.set_reveal_child(show_backward)

        # forward
        show_forward = self.current_page.can_navigate_forward or self.navigation.is_not_furthest()
        self.next_revealer.set_reveal_child(show_forward)

        # reload
        self.reload_revealer.set_reveal_child(self.current_page.can_reload)

    ### public methods ###

    def advance(self, page, allow_return: bool = True, cleanup: bool = False):
        if cleanup and allow_return:
            return print('Logic Error: Combining of return and cleanup not possible!')
        with self.navigation_lock:
            # confirm calling page is current page to prevent incorrect navigation
            if not page or page == self.current_page:
                if self.original_page_name:
                    if not allow_return:
                        return print('Logic Error: Returning unpreventable, page name mode')
                    self._load_original_page()
                else:
                    if not allow_return:
                        self.navigation.earliest = self.navigation.current + 1

                    self._load_page(self.navigation.current + 1)

                    if cleanup:
                        self._remove_pages(
                            self.pages[self.navigation.earliest:self.navigation.current - 1])

    def load_translated_pages(self):
        with self.navigation_lock:
            # delete pages that are not the language page
            self._remove_pages(self.pages[1:])
            self.pages = ['language']

            for page_name in self.available_pages.keys():
                if page_name != 'language':
                    self._initialize_page(page_name)

    def navigate_backward(self):
        with self.navigation_lock:
            if self.current_page.can_navigate_backward:
                self.current_page.navigate_backward()
            elif self.original_page_name:
                self._load_original_page()
            elif self.navigation.is_not_earliest():
                self._load_page(self.navigation.current - 1)

    def navigate_forward(self):
        with self.navigation_lock:
            if self.current_page.can_navigate_forward:
                self.current_page.navigate_forward()
            elif self.navigation.is_not_furthest():
                self._load_page(self.navigation.current + 1)

    def reload_page(self):
        with self.navigation_lock:
            if self.current_page.can_reload:
                self.current_page.load()

    def show_about_page(self):
        with self.navigation_lock:
            builder = Gtk.Builder.new_from_resource('/com/github/p3732/os-installer/ui/about_dialog.ui')
            popup = builder.get_object('about_window')
            if popup:
                self._show_dialog(popup)

    def show_confirm_quit_dialog(self, quit_callback):
        with self.navigation_lock:
            builder = Gtk.Builder.new_from_resource('/com/github/p3732/os-installer/ui/confirm_quit_popup.ui')
            popup = builder.get_object('popup')
            popup.connect('response',
                          lambda _, response: quit_callback() if response == "stop" else None)
            self._show_dialog(popup)

    def show_failed_page(self):
        with self.navigation_lock:
            global_state.installation_running = False

            self._load_page_by_name('failed')
            self.navigation.earliest = len(self.available_pages)
            self._update_navigation_buttons()

    def navigate_to_page(self, page_name):
        with self.navigation_lock:
            if self.original_page_name:
                return print('Logic Error: can only navigate by page name once in a row')
            self.original_page_name = self.main_stack.get_visible_child_name()
            self._load_page_by_name(page_name)



# SPDX-License-Identifier: GPL-3.0-or-later

from threading import Lock

from .global_state import global_state


class Preloadable:
    def __init__(self, preload_func):
        self.preload_func = preload_func
        self.preload_started = False
        self.preloaded = False
        self.preloading_lock = Lock()

    ### public methods ###

    def assert_preloaded(self):
        if self.preloaded:
            return

        with self.preloading_lock:
            if self.preloaded:
                return

            if not self.preload_started:
                print(f'Preloading for {self.__class__.__name__} was never started')
                self.preloading_lock.release()
                self.preload()
                self.preloading_lock.acquire()

            # await result
            self.future.result()
            self.preloaded = True

    def preload(self):
        with self.preloading_lock:
            if self.preload_started:
                return

            self.future = global_state.thread_pool.submit(self.preload_func)
            self.preload_started = True



# SPDX-License-Identifier: GPL-3.0-or-later

from concurrent.futures import ThreadPoolExecutor
import traceback
from .config import create_envs, init_config


class GlobalState:
    config = init_config()
    demo_mode = False
    installation_running = False

    thread_pool = ThreadPoolExecutor()  # for futures

    def __init__(self):
        self.set_config('disk_name', 'Test Dummy')

    def _uninitialized(self):
        print('Window method called before initialization.')
        traceback.print_stack()

    def get_config(self, setting):
        if setting in self.config:
            return self.config[setting]
        else:
            return None

    def set_config(self, setting, value):
        self.config[setting] = value

    def advance(self, *args):
        self._uninitialized()

    def load_translated_pages(self, *args):
        self._uninitialized()

    def navigate_to_page(self, *args):
        self._uninitialized()

    def reload_page(self, *args):
        self._uninitialized()

    def reload_title_image(self, *args):
        self._uninitialized()

    def send_notification(self, *args):
        self._uninitialized()

    def installation_failed(self, *args):
        self._uninitialized()

    def create_envs(self, with_install_envs=False, with_configure_envs=False):
        return create_envs(self.config, with_install_envs, with_configure_envs)


global_state = GlobalState()



# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Gtk

from .global_state import global_state
from .page import Page


@Gtk.Template(resource_path='/com/github/p3732/os-installer/ui/pages/user.ui')
class UserPage(Gtk.Box, Page):
    __gtype_name__ = __qualname__
    image = 'user-symbolic'

    user_name_row = Gtk.Template.Child()
    autologin_row = Gtk.Template.Child()
    password_row = Gtk.Template.Child()
    continue_button = Gtk.Template.Child()

    def __init__(self, **kwargs):
        Gtk.Box.__init__(self, **kwargs)

    def _set_continue_button(self):
        has_user_name = not self.user_name_row.get_text().strip() == ''
        autologin = self.autologin_row.get_active()
        has_password = not self.password_row.get_text() == ''
        can_continue = has_user_name and (autologin or has_password)
        self.continue_button.set_sensitive(can_continue)

    ### callbacks ###

    @Gtk.Template.Callback('autologin_row_clicked')
    def _autologin_row_clicked(self, row, state):
        self._set_continue_button()

    @Gtk.Template.Callback('focus_password')
    def _focus_password(self, row):
        self.password_row.grab_focus()

    @Gtk.Template.Callback('entry_changed')
    def _entry_changed(self, editable):
        self._set_continue_button()

    @Gtk.Template.Callback('continue')
    def _continue(self, object):
        if self.continue_button.get_sensitive():
            global_state.advance(self)

    ### public methods ###

    def unload(self):
        global_state.set_config('user_name', self.user_name_row.get_text().strip())
        global_state.set_config('user_password', self.password_row.get_text())
        global_state.set_config('user_autologin', self.autologin_row.get_active())



# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Gio, Gtk

from .global_state import global_state
from .installation_scripting import installation_scripting, Step
from .page import Page
from .widgets import reset_model, SummaryRow


@Gtk.Template(resource_path='/com/github/p3732/os-installer/ui/pages/summary.ui')
class SummaryPage(Gtk.Box, Page):
    __gtype_name__ = __qualname__
    image = 'checkbox-checked-symbolic'

    # rows
    language_row = Gtk.Template.Child()
    keyboard_row = Gtk.Template.Child()
    user_row = Gtk.Template.Child()
    format_row = Gtk.Template.Child()
    timezone_row = Gtk.Template.Child()
    software_row = Gtk.Template.Child()
    feature_row = Gtk.Template.Child()

    # row content
    user_label = Gtk.Template.Child()
    user_autologin = Gtk.Template.Child()

    # software list
    software_stack = Gtk.Template.Child()
    software_list = Gtk.Template.Child()
    software_model = Gio.ListStore()

    # feature list
    feature_stack = Gtk.Template.Child()
    feature_list = Gtk.Template.Child()
    feature_model = Gio.ListStore()

    def __init__(self, **kwargs):
        Gtk.Box.__init__(self, **kwargs)
        self.software_list.bind_model(
            self.software_model, lambda summary: SummaryRow(summary.name, summary.icon_path,
                                                            'application-x-executable-symbolic'))
        self.feature_list.bind_model(
            self.feature_model, lambda summary: SummaryRow(summary.name, summary.icon_path,
                                                       'puzzle-piece-symbolic'))
        self.language_row.set_visible(global_state.get_config('fixed_language'))
        self.software_row.set_visible(global_state.get_config('additional_software'))
        self.feature_row.set_visible(global_state.get_config('additional_features'))
        self.user_row.set_visible(not global_state.get_config('skip_user'))
        self.format_row.set_visible(not global_state.get_config('skip_locale'))
        self.timezone_row.set_visible(not global_state.get_config('skip_locale'))

    ### callbacks ###

    @Gtk.Template.Callback('continue')
    def _continue(self, button):
        global_state.advance(self, allow_return=False, cleanup=True)
        installation_scripting.set_ok_to_start_step(Step.configure)

    @Gtk.Template.Callback('summary_row_activated')
    def _summary_row_activated(self, list_box, row):
        global_state.navigate_to_page(row.get_name())

    ### public methods ###

    def load(self):
        self.language_row.set_subtitle(global_state.get_config('language'))
        self.keyboard_row.set_subtitle(
            global_state.get_config('keyboard_layout_ui'))
        self.user_label.set_label(global_state.get_config('user_name'))
        self.user_autologin.set_visible(
            global_state.get_config('user_autologin'))
        self.format_row.set_subtitle(global_state.get_config('formats_ui'))
        self.timezone_row.set_subtitle(global_state.get_config('timezone'))

        software = global_state.get_config('chosen_software')
        if len(software) > 0:
            self.software_stack.set_visible_child_name('used')
            reset_model(self.software_model, software)
        else:
            self.software_stack.set_visible_child_name('none')

        features = global_state.get_config('chosen_features')
        if len(features) > 0:
            self.feature_stack.set_visible_child_name('used')
            reset_model(self.feature_model, features)
        else:
            self.feature_stack.set_visible_child_name('none')

        return "prevent_back_navigation"



# SPDX-License-Identifier: GPL-3.0-or-later

import sys
from typing import Callable, NamedTuple

import gi
# set versions for all used submodules
gi.require_version('Gio', '2.0')           # noqa: E402
gi.require_version('GLib', '2.0')          # noqa: E402
gi.require_version('Gtk', '4.0')           # noqa: E402
gi.require_version('GnomeDesktop', '4.0')  # noqa: E402
gi.require_version('GWeather', '4.0')      # noqa: E402
gi.require_version('Adw', '1')             # noqa: E402
gi.require_version('Vte', '3.91')          # noqa: E402
from gi.repository import Adw, Gio, GLib, Gtk

# local, import order is important
from .global_state import global_state
from .preload_manager import preload_manager
from .window import OsInstallerWindow

APP_ID = 'com.github.p3732.OS-Installer'


class Action(NamedTuple):
    name: str
    func: Callable
    accel: list = []

class Application(Adw.Application):
    window = None

    def __init__(self, version):
        super().__init__(application_id=APP_ID,
                         flags=Gio.ApplicationFlags.HANDLES_COMMAND_LINE)

        # Connect app shutdown signal
        self.connect('shutdown', self._on_quit)

        # Additional command line options
        self.add_main_option('demo-mode', b'd', GLib.OptionFlags.NONE,
                             GLib.OptionArg.NONE, "Run in demo mode, don't alter the system", None)

        global_state.set_config('version', version)
        global_state.send_notification = self._send_notification
        preload_manager.start()

    def _setup_actions(self):
        actions = [
            Action('next-page', self._window('navigate_forward'), ['<Alt>Right']),
            Action('previous-page', self._window('navigate_backward'), ['<Alt>Left']),
            Action('reload-page', self._window('reload_page'), ['F5']),
            Action('about-page', self._window('show_about_page'), ['<Alt>Return']),
            Action('quit', self._on_quit, ['<Ctl>q']),
        ]

        for action in actions:
            self._add_action(action)

    def _add_action(self, action):
        gio_action = Gio.SimpleAction.new(action.name, None)
        gio_action.connect('activate', action.func)

        self.add_action(gio_action)

        if len(action.accel) > 0:
            self.set_accels_for_action('app.' + action.name, action.accel)

    def _setup_icons(self):
        icon_theme = Gtk.IconTheme.get_for_display(self.window.get_display())
        icon_theme.add_resource_path('/com/github/p3732/os-installer/')
        icon_theme.add_resource_path('/com/github/p3732/os-installer/icon')

    def _window(self, method_name):
        return (lambda _, __: getattr(self.window, method_name)())

    ### parent functions ###

    def do_activate(self):
        # create window if not existing
        if window := self.props.active_window:
            window.present()
        else:
            self.window = OsInstallerWindow(application=self)
            self._setup_icons()
            self.window.connect("close-request", self._on_quit)
            self.window.present()

            # load initial page
            self.window.advance(None)

    def do_command_line(self, command_line):
        options = command_line.get_options_dict()
        options = options.end().unpack()

        if 'demo-mode' in options:
            global_state.demo_mode = True
            self._add_action(Action('fail-page', self._window('show_failed_page'), ['<Alt>F']))

        self.activate()
        return 0

    def do_startup(self):
        # Startup application
        self.set_resource_base_path('/com/github/p3732/os-installer')
        Adw.Application.do_startup(self)
        self._setup_actions()

    ### callbacks ###

    def _on_quit(self, action, param=None):
        if global_state.installation_running:
            # show confirm dialog
            self.window.show_confirm_quit_dialog(self.quit)
            # return True to avoid further processing of event
            return True
        else:
            self.quit()

    def _send_notification(self, title, text):
        n = Gio.Notification()
        n.set_title(title)
        n.set_body(text)
        self.send_notification(APP_ID, n)

def main(version):
    app = Application(version)
    return app.run(sys.argv)



# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Gtk

from .global_state import global_state
from .page import Page


@Gtk.Template(resource_path='/com/github/p3732/os-installer/ui/pages/done.ui')
class DonePage(Gtk.Box, Page):
    __gtype_name__ = __qualname__
    image = 'success-symbolic'

    page_title = Gtk.Template.Child()

    def __init__(self, **kwargs):
        Gtk.Box.__init__(self, **kwargs)

    ### callbacks ###

    @Gtk.Template.Callback('restart_button_clicked')
    def _restart_button_clicked(self, button):
        global_state.advance(self, allow_return=False)

    ### public methods ###

    def load_once(self):
        global_state.send_notification(self.page_title.get_label(), '')



# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import GObject
from gi.repository.GnomeDesktop import XkbInfo

# generated via language_codes_to_x_generator.py
language_to_default_keyboard = {
    'ab': 'ru', 'aa': 'us', 'af': 'us+intl', 'ak': 'us+altgr-intl', 'sq': 'al', 'am': 'et', 'ar': 'ara', 'an': 'es',
    'hy': 'am', 'as': 'in+eng', 'av': 'ru', 'ae_AE': 'ara', 'ay': 'latam', 'az': 'az', 'bm': 'us', 'ba': 'ru+bak',
    'eu': 'es', 'be': 'by', 'bn': 'in+eng', 'bh_BH': 'ara', 'bi': 'us+euro', 'bs': 'ba', 'br': 'fr+bre', 'bg': 'bg',
    'my': 'mm', 'ca': 'es+cat', 'ch': 'us', 'ce': 'ru', 'ny': 'us', 'zh': 'cn', 'cv': 'ru+cv', 'kw': 'gb',
    'co': 'fr+oss', 'cr_CR': 'latam', 'hr': 'hr', 'cs': 'cz', 'da': 'dk', 'dv': 'mv', 'nl': 'us+euro', 'dz': 'bt',
    'en': 'us', 'en_GB': 'gb', 'eo': 'us+altgr-intl', 'et': 'ee', 'ee': 'us', 'fo': 'fo', 'fj': 'us', 'fi': 'fi',
    'fr': 'fr+oss', 'ff': 'fr+oss', 'gl': 'es', 'ka': 'ge', 'de': 'de+nodeadkeys', 'el': 'gr', 'gn': 'us',
    'gu': 'in+eng', 'ht': 'latam', 'ha': 'ng+hausa', 'he': 'il', 'hz': 'us', 'hi': 'in+eng', 'ho': 'us',
    'hu': 'hu', 'ia': 'us+euro', 'id': 'us', 'ie': 'us', 'ga': 'ie+CloGaelach', 'ig': 'ng+igbo', 'ik': 'ca+eng',
    'io': 'us', 'is': 'is', 'it': 'it', 'iu': 'ca+ike', 'ja': 'jp', 'jv': 'us', 'kl': 'us', 'kn': 'in+eng',
    'kr': 'us', 'ks': 'in+eng', 'kk': 'kz', 'km': 'kh', 'ki': 'us', 'rw': 'us', 'ky': 'kg', 'kv': 'ru+kom',
    'kg': 'fr+oss', 'ko': 'kr', 'ku': 'tr+ku', 'kj': 'us', 'la': 'it', 'lb': 'fr+oss', 'lg': 'us', 'li': 'us+euro',
    'lo': 'la', 'lt': 'lt', 'lu': 'us', 'lv': 'lv', 'gv': 'gb', 'mk': 'mk', 'mg': 'us', 'ms': 'us', 'ml': 'in+eng',
    'mt': 'mt', 'mi': 'us', 'mr': 'in+eng', 'mh': 'us', 'mn': 'mn', 'na': 'us', 'nv': 'us', 'nd': 'us', 'ne': 'np',
    'ng': 'us', 'nb': 'no', 'nn': 'no', 'no': 'no', 'ii': 'us', 'nr': 'us', 'oc': 'fr+oss', 'cu': 'bg', 'om': 'us',
    'or': 'in+eng', 'os': 'ru+os_winkeys', 'pa': 'in+eng', 'fa': 'ir', 'pl': 'pl', 'ps': 'af+ps', 'pt': 'br',
    'qu': 'latam', 'rm': 'ch', 'rn': 'us', 'ro': 'ro', 'ru': 'ru', 'sa': 'in+eng', 'sc': 'it', 'sd': 'in+eng',
    'se': 'no', 'sm': 'us+euro', 'sg': 'us', 'sr': 'rs', 'gd': 'gb', 'sn': 'us', 'si': 'us', 'sk': 'sk', 'sl': 'si',
    'so': 'us', 'st': 'us', 'es': 'es', 'su': 'us', 'sw': 'ke', 'ss': 'us', 'sv': 'se', 'ta': 'in+eng', 'te': 'in+eng',
    'tg': 'tj', 'th': 'th', 'ti': 'et', 'bo': 'us', 'tk': 'tm', 'tl': 'ph', 'tn': 'za', 'to': 'us', 'tr': 'tr',
    'ts': 'us', 'tt': 'ru+tt', 'tw': 'us', 'ty': 'fr+oss', 'ug': 'cn+ug', 'uk': 'ua', 'ur': 'pk', 'uz': 'uz',
    've': 'za', 'vi': 'vn', 'vo': 'de+nodeadkeys', 'wa': 'be+oss', 'cy': 'gb', 'wo': 'sn', 'fy': 'us+euro',
    'xh': 'us', 'yi': 'us', 'yo': 'ng+yoruba', 'za': 'us', 'zu': 'us',
    # manually added
    'pt_PT': 'pt',
}


fallback_codes = {
    'oc': 'fr'
}

xkb_info = XkbInfo()

class KeyboardInfo(GObject.GObject):
    __gtype_name__ = __qualname__
    name: str
    layout: str

    def __init__(self, name, layout):
        super().__init__()

        self.name = name
        self.layout = layout


def _fallback_code(language_code):
    if language_code in fallback_codes:
        return fallback_codes[language_code]


def _short_code(code):
    return code.split('_')[0]


def _get_existing_layouts(language_code):
    xkb_info = XkbInfo()
    layouts = xkb_info.get_layouts_for_language(language_code)
    if len(layouts) > 0:
        return layouts

    if not (short_code := _short_code(language_code)) == language_code:
        layouts = xkb_info.get_layouts_for_language(short_code)
        if len(layouts) > 0:
            return layouts

    if fallback_code := _fallback_code(language_code):
        layouts = xkb_info.get_layouts_for_language(fallback_code)
        return layouts
    else:
        print(f'Language {language_code} has no keyboard layouts! Please report this.')
        return ['us']


def _get_default_layout_code(language_code):
    if language_code in language_to_default_keyboard:
        return language_to_default_keyboard[language_code]
    elif (short_code := _short_code(language_code)) in language_to_default_keyboard:
        return language_to_default_keyboard[short_code]
    elif (fallback_code := _fallback_code(language_code)) in language_to_default_keyboard:
        return language_to_default_keyboard[fallback_code]
    else:
        return 'us'

def get_default_layout(language_code):
    layout_code = _get_default_layout_code(language_code)
    name = xkb_info.get_layout_info(layout_code).display_name
    return KeyboardInfo(name, layout_code)

def get_layouts_for(language_code, language):
    layouts = _get_existing_layouts(language_code)

    named_layouts = []
    for layout in layouts:
        name = xkb_info.get_layout_info(layout).display_name
        named_layouts.append(KeyboardInfo(name, layout))

    default_layout = _get_default_layout_code(language_code)
    # Sort the layouts, prefer those starting with language name or matching language short hand. Then by name.
    return sorted(named_layouts, key=lambda o:
                  (not o.layout == default_layout,
                   not o.name.startswith(language),
                   o.name))



# SPDX-License-Identifier: GPL-3.0-or-later

from enum import Enum
from gi.repository import Gtk

from .format_provider import get_formats
from .global_state import global_state
from .page import Page
from .system_calls import set_system_formats, set_system_timezone
from .timezone_provider import get_timezones
from .widgets import reset_model, ProgressRow


class FilterType(Enum):
    format = 0
    timezone = 1


@Gtk.Template(resource_path='/com/github/p3732/os-installer/ui/pages/filter.ui')
class FilterPage(Gtk.Box, Page):
    __gtype_name__ = __qualname__
    image = 'map-symbolic'

    format_title = Gtk.Template.Child()
    timezone_title = Gtk.Template.Child()

    search_entry = Gtk.Template.Child()
    custom_filter = Gtk.Template.Child()
    filter_list_model = Gtk.Template.Child()

    stack = Gtk.Template.Child()
    list = Gtk.Template.Child()
    list_loaded = False
    list_model = Gtk.Template.Child()

    def __init__(self, filter_type, **kwargs):
        Gtk.Box.__init__(self, **kwargs)

        self.type = filter_type
        match self.type:
            case FilterType.format:
                self.format_title.set_visible(True)
                self.filter = self._format_filter
            case FilterType.timezone:
                self.timezone_title.set_visible(True)
                self.filter = self._timezone_filter

        self.search_entry.connect("search-changed", self._filter)

        self.list.bind_model(
            self.filter_list_model, lambda f: ProgressRow(f.name))

    def _filter(self, *args):
        self.search_text = self.search_entry.get_text().lower()
        self.custom_filter.set_filter_func(self.filter)

        if self.filter_list_model.get_n_items() > 0:
            self.stack.set_visible_child_name('list')
        else:
            self.stack.set_visible_child_name('none')

    def _format_filter(self, format):
        return self.search_text in format.lower_case_name or format.locale.startswith(self.search_text)

    def _timezone_filter(self, timezone):
        if self.search_text in timezone.lower_case_name:
            return True
        for location in timezone.locations:
            if self.search_text in location:
                return True
        return False

    ### callbacks ###

    @Gtk.Template.Callback('row_selected')
    def _row_selected(self, list_box, row):
        match self.type:
            case FilterType.format:
                set_system_formats(row.info, row.get_label())
            case FilterType.timezone:
                set_system_timezone(row.get_label())
        global_state.advance(self)

    ### public methods ###

    def load(self):
        if not self.list_loaded:
            self.list_loaded = True
            match self.type:
                case FilterType.format:
                    model = get_formats()
                case FilterType.timezone:
                    model = get_timezones()
            reset_model(self.list_model, model)
        return "load_next"


FormatPage = lambda **args: FilterPage(FilterType.format, **args)
TimezonePage = lambda **args: FilterPage(FilterType.timezone, **args)



# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Gtk

from .global_state import global_state
from .installation_scripting import installation_scripting
from .page import Page


@Gtk.Template(resource_path='/com/github/p3732/os-installer/ui/pages/install.ui')
class InstallPage(Gtk.Box, Page):
    __gtype_name__ = __qualname__
    image = 'OS-Installer-symbolic'

    terminal_box = Gtk.Template.Child()
    stack = Gtk.Template.Child()
    spinner = Gtk.Template.Child()

    def __init__(self, **kwargs):
        Gtk.Box.__init__(self, **kwargs)

        installation_scripting.install_page_name = self.__gtype_name__

        # UI element states
        self.stack.set_visible_child_name('spinner')

    ### callbacks ###

    @Gtk.Template.Callback('terminal_button_toggled')
    def _terminal_button_toggled(self, toggle_button):
        if self.stack.get_visible_child_name() == 'spinner':
            self.spinner.stop()
            self.stack.set_visible_child_name('terminal')
        else:
            self.spinner.start()
            self.stack.set_visible_child_name('spinner')

    ### public methods ###

    def load(self):
        # setup terminal
        self.terminal_box.append(installation_scripting.terminal)
        self.spinner.start()

    def unload(self):
        self.terminal_box.remove(installation_scripting.terminal)
        self.spinner.stop()



# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Gtk

from .system_calls import reboot_system
from .page import Page


@Gtk.Template(resource_path='/com/github/p3732/os-installer/ui/pages/restart.ui')
class RestartPage(Gtk.Box, Page):
    __gtype_name__ = __qualname__
    image = 'system-reboot-symbolic'

    spinner = Gtk.Template.Child()

    def __init__(self, **kwargs):
        Gtk.Box.__init__(self, **kwargs)

    ### public methods ###

    def load(self):
        self.spinner.start()
        reboot_system()



# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Gtk

from .global_state import global_state
from .page import Page


@Gtk.Template(resource_path='/com/github/p3732/os-installer/ui/pages/encrypt.ui')
class EncryptPage(Gtk.Box, Page):
    __gtype_name__ = __qualname__
    image = 'dialog-password-symbolic'

    switch_row = Gtk.Template.Child()
    pin_row = Gtk.Template.Child()
    info_revealer = Gtk.Template.Child()

    continue_button = Gtk.Template.Child()

    def __init__(self, **kwargs):
        Gtk.Box.__init__(self, **kwargs)

    def _set_continue_button(self):
        needs_pin = self.switch_row.get_active()
        has_pin = len(self.pin_row.get_text()) > 0
        self.continue_button.set_sensitive(not needs_pin or has_pin)

    ### callbacks ###

    @Gtk.Template.Callback('encryption_row_clicked')
    def _encryption_row_clicked(self, row, state):
        state = self.switch_row.get_active()
        self.pin_row.set_sensitive(state)
        self.info_revealer.set_reveal_child(state)
        self._set_continue_button()

        if state:
            self.pin_row.grab_focus()

    @Gtk.Template.Callback('pin_changed')
    def _pin_changed(self, editable):
        self._set_continue_button()

    @Gtk.Template.Callback('continue')
    def _continue(self, object):
        if self.continue_button.is_sensitive():
            global_state.advance(self)

    ### public methods ###

    def unload(self):
        use_encryption = self.switch_row.get_active()
        pin = self.pin_row.get_text()
        global_state.set_config('use_encryption', use_encryption)
        global_state.set_config('encryption_pin', pin)



# SPDX-License-Identifier: GPL-3.0-or-later

from enum import Enum
from gi.repository import Gio, Gtk

from .choices_provider import choices_provider
from .global_state import global_state
from .page import Page
from .util import SummaryEntry
from .widgets import MultiSelectionRow, reset_model, SelectionRow


class ChoiceType(Enum):
    feature = 0
    software = 1


@Gtk.Template(resource_path='/com/github/p3732/os-installer/ui/pages/choices.ui')
class ChoicesPage(Gtk.Box, Page):
    __gtype_name__ = __qualname__

    choices_list = Gtk.Template.Child()
    software_header = Gtk.Template.Child()
    feature_header = Gtk.Template.Child()
    model = Gtk.Template.Child()

    def __init__(self, choice_type, **kwargs):
        Gtk.Box.__init__(self, **kwargs)

        self.type = choice_type
        match self.type:
            case ChoiceType.feature:
                self.image = 'puzzle-piece-symbolic'
                self.feature_header.set_visible(True)
                self.list_provider = choices_provider.get_feature_suggestions
            case ChoiceType.software:
                self.image = 'system-software-install-symbolic'
                self.software_header.set_visible(True)
                self.list_provider = choices_provider.get_software_suggestions
            case _:
                print("Unknown choice type!")
                exit(0)

        self.choices_list.bind_model(
            self.model,
            self._create_row)

    def _create_row(self, choice):
        if choice.options:
            return MultiSelectionRow(choice.name, choice.description, choice.icon_path,
                                     'application-x-executable-symbolic',
                                     choice.options)
        else:
            return SelectionRow(choice.name, choice.description, choice.icon_path,
                                'application-x-executable-symbolic',
                                choice.suggested, choice.keyword)

    ### callbacks ###

    @Gtk.Template.Callback('continue')
    def _continue(self, button):
        global_state.advance(self)

    @Gtk.Template.Callback('row_activated')
    def _row_activated(self, list_box, row):
        if type(row) == SelectionRow:
            row.flip_switch()

    ### public methods ###

    def load_once(self):
        reset_model(self.model, self.list_provider())

    def unload(self):
        summary = []
        keywords = []
        for row in self.choices_list:
            if type(row) == MultiSelectionRow:
                option = row.get_chosen_option()
                keywords.append(option.keyword)
                summary.append(SummaryEntry(option.display, row.icon_path))
            elif type(row) == SelectionRow and row.is_activated():
                keywords.append(row.info)
                summary.append(SummaryEntry(row.get_title(), row.icon_path))
        keywords = ' '.join(keywords)
        match self.type:
            case ChoiceType.feature:
                global_state.set_config('chosen_feature_names', keywords)
                global_state.set_config('chosen_features', summary)
            case ChoiceType.software:
                global_state.set_config('chosen_software_packages', keywords)
                global_state.set_config('chosen_software', summary)


FeaturePage = lambda **args: ChoicesPage(ChoiceType.feature, **args)
SoftwarePage = lambda **args: ChoicesPage(ChoiceType.software, **args)
